/***************************************************************************
 *   Copyright (C) 2007 by Andrea Moro   *
 *   andrea_moro@email.it   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program.  If not, see <http://www.gnu.org/licenses/>. *
 ***************************************************************************/
#ifndef HILL_H
#define HILL_H

#include "ui_hill.h"

class mainWindow;
class QLineEdit;

/** 
 * @short User interface and algorithms for Hill cipher.
 *
 * The Hill cipher. This class provides the cryptographic algorithms
 * and user interface handling.
 */
class hill :
		public QWidget,
		public Ui::hill
{
	Q_OBJECT
public:
	/**
	 * Creates a new Hill instance.
	 *
	 * @param *main the main window of CrypTooLinux
	 * @param text the text to encrypt or decrypt
	 * @param title the title of the window taht contain the text to encrypt or decrypt
	 */
	hill(mainWindow *main,QString text, QString title);
private:
	/**
	 * An enumeration used to tell the ciph() function whether we actually
	 * encrypt or decrypt the text.
	 * 
	 */
	enum ciphFlag {ENCIPHER, DECIPHER};
	/**
	 * Draw a Hill window with two 5x5 matrix.
	 */
	void littleHill();
	/**
	 * Draw a Hill window with two 10x10 matrix.
	 */
	void bigHill();
	/**
	 * Instead of just make the encrypt or decrypt window add another window
	 * with an explaination of the hill method.
	 */
	void explain();
	/**
	 * Calculate the modular inverse of a matrix. If is it possible to calculate
	 * the inverse and if the second parameter is equal to 1 the inverse will be
	 * in the first parameter.
	 *
	 * @param matrix the start matrix and, if is it possible to calculate
	 * the inverse and if the second parameter is equal to 1, the inverse
	 * @param modify if is equal to 0 matrix is unchanged, else in matrix
	 * there will be the inverse
	 * @return true if it is possible to calculate the inverse, else false
	 * @bug it cannot calculate all the possible modular inverse
	 */
	bool inverse(int** matrix,int modify);
	/**
	 * The Extended Euclidean algorithm(http://en.wikipedia.org/wiki/Extended_Euclidean_algorithm)
	 * Calculate the greatest common divisor and the multiplicative inverse of the two parameters.
	 *
	 * @return a vector of three elements, the GCD, the inverse of a in Z_b,
	 * the inverse of b in Z_a
	 */
	int* ext_gcd(int a,int b);
	/**
	 * Check if the matrix is valid.
	 *
	 * @return true if the matrix is valid, else false
	 */
	bool checkKey();
	/**
	 * The dimension of the square matrix
	 */
	int kLen;
	/**
	 * big==true if the window shows 2 10x10 matrices, big==false if the window shows 2 5x5 matrices
	 */
	bool big;
	/**
	 * Makes the connections between signal of the ui and the slot functions.
	 */
	void createConnections();
	/**
	 * The text to encrypt or decrypt.
	 */
	QString workText;
	/**
	 * The title of window in CrypTooLinux main window that contain the text
	 * to encrypt or decrypt.
	 */
	QString workTitle;
	/**
	 * CrypTooLinux main window.
	 */
	mainWindow *mw;

	/*graphics objects*/

	QLineEdit *cellA[10][10];
	QLineEdit *cellN[10][10];
	QGroupBox *descBox;
	QGroupBox *alphBox;
	QGroupBox *matrixBox;
	QGroupBox *moltBox;
	QGroupBox *sizeBox;
	QGroupBox *alphMatrix;
	QGroupBox *numbMatrix;
	QLabel *descLabel;
	QLineEdit *alph;
	QRadioButton *alphM;
	QRadioButton *numbM;
	QPushButton *rndMatrix;
	QRadioButton *variant1;
	QRadioButton *variant2;
	QRadioButton *sizeRadio[10];
	QPushButton *largerMatrix;
	QCheckBox *explainCheck;
	QPushButton *encButton;
	QPushButton *decButton;
	QPushButton *textOpt;
	QPushButton *cancButton;


private slots:
	/**
	 * When an element in a matrix is changed or deleted check if the value is correct
	 * and update the respective value in the other matrix
	 */
	void upMatr();
	/**
	 * Generate a random valid matrix.
	 */
	void randKey();
	/**
	 * Change the dimension of active matrix.
	 */
	void changeDim();
	/**
	 * Decrypt the workText with the matrix inserted, close the Hill window and create
	 * a new window in CrypTooLinux main window with the decrypted text.
	 */
	void decrypt();
	/**
	 * Encrypt the workText with the matrix inserted, close the Hill window and create
	 * a new window in CrypTooLinux main window with the encrypted text.
	 */
	void encrypt();
	/**
	 * If the current window contain two 10x10 matrix then call littleHill()
	 * else if the current window contain two 5x5 matrix then call bigHill()
	 */
	void changeWin();
	/**
	 * In case the alphabet is changed(with text options) update the alphabet
	 * to the new alphabet
	 */
	void updAlph();

	/**
	 * Properly encrypt or decrypt the text, with the inserted
	 * matrix. Closes the Hill window and creates a new window in
	 * CrypTooLinux main window with the ciphertext/deceiphered text.
	 * 
	 * @param flag a flag indicating whether we shall encrypt or decrypt.
	 */
	void ciph(const ciphFlag& flag);
	
};
#endif
