/***************************************************************************
 *   Copyright (C) 2007 by apterium   *
 *   apterium@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __LIB_ANALYSIS_CUSTOM_FACTORISATION_H
#define __LIB_ANALYSIS_CUSTOM_FACTORISATION_H

#include <list>
#include "gmpxx.h"
#include "Factorisation.h"
#include "FactorisationBruteforce.h"
#include "FactorisationFermat.h"

namespace libAnalysis
{

	/**
	 * Enumeration of the selected algorithms. Used to record which
	 * algorithm the user wants us to use, thus allowing to switch
	 * from an
	 * algorithm to an other one if it is faster at the point the
	 * computation has reached.
	 * For example : some algorithms have to factor
	 * smaller numbers in order to factor the previous one. Then
	 * relying on
	 * the number size, the program should be able to choose the
	 * optimal
	 * algorithm to factor the number. Especially if it is
	 * multithreaded, we
	 * may try fast algorithm when they work (for example ECM and
	 * Pollards
	 * Rho method) and slower one, but giving a result faster if
	 * previous
	 * ones fail, in parallel.
	 *
	 */
	enum FactorisationAlgorithmIndex {
		BRUTEFORCE ,
		FERMAT /* , POLLARD-RHO, etc. */
	};
	



	/**
	 * Class designed to allow the user to select which factoring algorithms
	 * he wants to allow, and to customise the design of the sequential use
	 * of many algorithms. For example, when a factor has been  extracted,
	 * the result is to be factored, using perhaps an other, faster
	 * algorithm. It may, depending on the subnumber size, 
	 * factor it using an in practice faster algorithm which is not
	 * necessarily the initial factorign algorithm.
	 * 
	 * Recommended use is when you need a factoring algorithm with selected
	 * unused ones. For example, the user specifies which algorithms are to
	 * be used. Or for a generic implementation, an algorithm will use
	 * subalgorithms to factor new computed integers.
	 *
	 * @author apterium
	 */
	class CustomFactorisation : public Factorisation
	{
	public:		
	
		/**
		 * Creates a new CustomFactorisation instance. By default, every
		 * algorithms will be allowed.
		 *
		 * @param n the number to factorise.
		 */
		CustomFactorisation(const mpz_class& n);
		
		/**
		 * Creates a new CustomFactorisation instance, given the allowed algorithms.
		 *
		 * @param n the number to be factorised.
		 * @param l the list of the allowed algorithms. That is, members
		 * of the upper enumeration
		 */
		CustomFactorisation(const mpz_class& n, const std::list<FactorisationAlgorithmIndex>& l);

		/**
		 * Destructor.
		 *
		 */
		~CustomFactorisation();

		/**
		 * @todo do NOT hardcode the primality test.
		 * @see Factorisation.
		 * 
		 */
		mpz_class getNextFactor() throw (IsPrimeException);

		/**
		 * Return whether the number we are attempting to factor is
		 * prime or not. To be used in order to avoid the unnecessary
		 * use of exceptions if the number is easily detected as prime.
		 *
		 *
		 * @todo use the same type of custom primality tests
		 * implementation in libNumberTheory and use it here instead of
		 * hardcoding the use of GMP test.
		 * @todo do NOT use exception. The next factor of a prime is
		 * 1. Who wants to factor a number without knowing it's not
		 * prime ? The caller MUST HIMSELF test whether the number is
		 * prime or not, which is much faster than trying to factor it
		 * if it's prime.
		 * @return 0 if it is definitely composite, 1 if it is probably
		 * prime, 2 if it is definitely prime.
		 */
		int isPrime();

		/**
		 * @see Factorisation
		 * 
		 */
		mpz_class getRemainingQuotient();


	private:
		/**
		 * The list of allowed algorithms, among which we pick up the engine.
		 * 
		 */
		std::list<FactorisationAlgorithmIndex> *indexes;


		/**
		 * ?
		 *
		 */
		Factorisation *engine;

		/**
		 * A private wrapper around the desired primality testing method.
		 * 
		 * @see CompositenessTest.h
		 */
		int primalityTest();
		
	};

} //namespace libAnalysis

#endif
