/*************************************************************************** 
 *   Copyright (C) 2007 by apterium   *                                      
 *   apterium@gmail.com   *                                                  
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  * 
 *   it under the terms of the GNU General Public License as published by  * 
 *   the Free Software Foundation; either version 3 of the License, or     * 
 *   (at your option) any later version.                                   * 
 *                                                                         * 
 *   This program is distributed in the hope that it will be useful,       * 
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        * 
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         * 
 *   GNU General Public License for more details.                          * 
 *                                                                         * 
 *   You should have received a copy of the GNU General Public License     * 
 *   along with this program; if not, write to the                         * 
 *   Free Software Foundation, Inc.,                                       * 
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             * 
 ***************************************************************************/

#ifndef __LIB_CLASSICAL_CRYPTO_CIPHER_ENGINE_H
#define __LIB_CLASSICAL_CRYPTO_CIPHER_ENGINE_H

#include <vector>
#include <gmpxx.h>
#include "../libalgebra/libalgebra.h"

namespace LibClassicalCrypto
{
	/**
	 * @short Interface providing key insertion, key validation, and cipher/decipher
	 * operations on a mpz_class vector representing the text, including
	 * basic translations functions.
	 *
	 * This class is only an API every crypto engine should implement.
	 * 
	 * 
	 * @author apterium 
	 */
	class CipherEngine
	{
	public:
		/**
		 * Destructor. Virtual as this class is an interface.
		 *
		 */
		~CipherEngine();

		/**
		 * Computes the cipher text of a given plaintext, under the
		 * inserted key.
		 *  
		 * @param plainText a constant reference to the vector
		 * representing the plaintext.
		 * @return the vector representing the ciphertext of the
		 * plaintext with the selected algorithm and the inserted key.
		 */ 
		virtual std::vector<mpz_class> encipher(const std::vector<mpz_class>& plainText) const = 0;

		/**
		 * Computes the plaintext of the given ciphertext with the
		 * specified key.
		 * 
		 * @param cipherText a constant reference to the vector
		 * representing the ciphertext.
		 * @return the vector representing the plaintext.
		 */
		virtual std::vector<mpz_class> decipher(const std::vector<mpz_class>& cipherText) const = 0;
		
		/**
		 * Translates a text into the corresponsding vector of RingZnElement.
		 * 
		 * @param text a vector representing the text, each element
		 * represents the position of the letter in the alphabet.
		 * @param ring the RingZn the elements will belong to.
		 * @return the vector of RingZnElement representing the text.
		 */
		std::vector<libalgebra::RingZnElement> toRingZnElement(const std::vector<mpz_class>& text, 
								       const libalgebra::RingZn* ring) const;
		
		/**
		 * Inverse of toRingZnElement, that is compute the text under
		 * the form of a vector of mpz_class from a text under the form
		 * of a vector of RingZnElements.
		 * 
		 * @param text a constant reference to the text
		 * @return the text under the form of a vector of
		 * mpz_class. Directly usable in the Text class of CrypTooLinux.
		 */
		std::vector<mpz_class> toMpz_class(const std::vector<libalgebra::RingZnElement>& text) const;

	};

} //namespace LibClassicalCrypto

#endif //__LIB_CLASSICAL_CRYPTO_TEXT_H
