/*************************************************************************** 
 *   Copyright (C) 2007 by apterium   *                                      
 *   apterium@gmail.com   *                                                  
 *                                                                         * 
 *   This program is free software; you can redistribute it and/or modify  * 
 *   it under the terms of the GNU General Public License as published by  * 
 *   the Free Software Foundation; either version 3 of the License, or     * 
 *   (at your option) any later version.                                   * 
 *                                                                         * 
 *   This program is distributed in the hope that it will be useful,       * 
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        * 
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         * 
 *   GNU General Public License for more details.                          * 
 *                                                                         * 
 *   You should have received a copy of the GNU General Public License     * 
 *   along with this program; if not, write to the                         * 
 *   Free Software Foundation, Inc.,                                       * 
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             * 
 ***************************************************************************/

#include "HillEngine.h"

namespace LibClassicalCrypto
{
	HillEngine::HillEngine(int** matrix, const mpz_class& modulus, int dim)
		throw(InvalidKeyException)
	{
		this->ring = new libalgebra::RingZn(modulus);
		std::vector<libalgebra::RingZnElement> row(dim, libalgebra::RingZnElement(this->ring) );
		std::vector<std::vector<libalgebra::RingZnElement> > matr(dim, row);
		for (int i = 0 ; i < dim ; i++) {
			for (int j = 0 ; j < dim ; j++) {
				matr.at(i).at(j) = 
					libalgebra::RingZnElement(mpz_class(matrix[i][j]), this->ring);
			}
		}
		this->key = new 
			libalgebra::DenseMatrix<libalgebra::RingZnElement, libalgebra::RingZn>
			(matr, this->ring);
		try {
			this->invKey = new libalgebra::DenseMatrix<libalgebra::RingZnElement,
				libalgebra::RingZn>(this->key->inverse() );
		} 
		catch (libalgebra::NonInvertibleMatrixException& exc) {
			delete this->key;
			delete this->ring;
			throw (InvalidKeyException());
		}
		this->dimension = dim;
		
	    #ifdef DEBUG
		std::cout << "Encryption key : " << *(this->key) << std::endl;
		std::cout << "Decryption key : " << *(this->invKey) << std::endl;
            #endif

	}
	
	HillEngine::HillEngine()
	{
		this->key = NULL;
		this->ring = NULL;
		this->invKey = NULL;
	}

	HillEngine::HillEngine(const HillEngine& copy) : CipherEngine(copy)
	{
		this->ring = new libalgebra::RingZn(copy.ring->getModulus());
		int dim = copy.key->getRowDimension().get_ui();
		//then copy the values : each engine has its own ring...
		std::vector<libalgebra::RingZnElement> row(dim, libalgebra::RingZnElement(this->ring) );
		std::vector<std::vector<libalgebra::RingZnElement> > matr(dim, row);
		for (std::vector<std::vector<libalgebra::RingZnElement> >::size_type i = 0 ; i < row.size() ; i++) {
			for (std::vector<libalgebra::RingZnElement>::size_type j = 0 ; j < row.size() ; j++) {
				matr.at(i).at(j) = 
					libalgebra::RingZnElement(copy.key->getElement(mpz_class(i), mpz_class(j)).getValue(), this->ring);
			}
		}
		this->key = new
			libalgebra::DenseMatrix<libalgebra::RingZnElement, libalgebra::RingZn>
			(matr, this->ring);
		this->invKey = new
			libalgebra::DenseMatrix<libalgebra::RingZnElement, libalgebra::RingZn>
			(this->key->inverse());
	}
	
	HillEngine::~HillEngine()
	{
		delete this->key;
		delete this->invKey;
		delete this->ring;
	}
	
	void HillEngine::getInverseKey(int** matrix, int length)
		throw(libalgebra::DimensionMismatchException)
	{
		if (this->dimension != length) {
			throw (libalgebra::DimensionMismatchException());
		}
		unsigned int l = length;
		for (std::vector<std::vector<libalgebra::RingZnElement> >::size_type i = 0 ; i < l ; i++) {
			for (std::vector<libalgebra::RingZnElement>::size_type j = 0 ; j < l ; j++) {
				matrix[i][j] = (int) this->invKey->getElement(mpz_class(i + 1), mpz_class(j + 1)).getPositiveValue().get_si();
			}
		}
		
		
	}
	
	std::vector<mpz_class> HillEngine::encipher(const std::vector<mpz_class>& plainText) const
	{
		std::vector<libalgebra::RingZnElement> cipherText
			= this->compute(this->toRingZnElement(plainText, this->ring), *(this->key) );
		
		return this->toMpz_class(cipherText);
	}
	
	std::vector<mpz_class> HillEngine::decipher(const std::vector<mpz_class>& cipherText) const
	{
		std::vector<libalgebra::RingZnElement> plainText =
			this->compute(this->toRingZnElement(cipherText, this->ring), *(this->invKey) );
		return this->toMpz_class(plainText);
	}

	std::ostream& operator<<(std::ostream& out, const HillEngine& engine)
	{
		out << engine.ring->getModulus() << std::endl;
		out << *(engine.key) << std::endl;
		return out;
	}

	std::vector<libalgebra::RingZnElement> HillEngine::compute(const std::vector<libalgebra::RingZnElement>& text,
								   const libalgebra::DenseMatrix<libalgebra::RingZnElement, libalgebra::RingZn>& matrix) const
	{
		std::vector<libalgebra::RingZnElement> ret;
		ret.reserve(text.size());

		typedef std::vector<libalgebra::RingZnElement>::size_type size_type;

		size_type dim = (size_type) this->dimension;
		size_type blocksNbr = text.size() / dim;		
		
		#ifdef DEBUG
		std::cout << "The number of blocks : " << blocksNbr << std::endl;
		#endif
		
		
		for (size_type i = 0 ; i < blocksNbr ; i++) {
			std::vector<libalgebra::RingZnElement> block;
			block.reserve(dim);
			//copy the text into the block. Should be improved ?
			for (size_type j = 0 ; j < dim ; j++) {
				block.push_back(text.at(i * dim + j) );
			}
			
			std::vector<libalgebra::RingZnElement> resultBlock 
				= this->proceedBlock(block, matrix);


			//copy the resulting block into the return vector.
			for (size_type j = 0 ; j < dim ; j++) {
					ret.push_back(resultBlock.at(j));
			}
		}
		return ret;
	}

	std::vector<libalgebra::RingZnElement> HillEngine::proceedBlock(const std::vector<libalgebra::RingZnElement>& textBlock,
		const libalgebra::DenseMatrix<libalgebra::RingZnElement, libalgebra::RingZn>& matrix) const
	{
		libalgebra::DenseMatrix<libalgebra::RingZnElement, libalgebra::RingZn> matTxt = 
			libalgebra::DenseMatrix<libalgebra::RingZnElement, libalgebra::RingZn>::columnVector(textBlock, this->ring);
		
		libalgebra::DenseMatrix<libalgebra::RingZnElement, libalgebra::RingZn> resTxt = matrix * matTxt;
		return resTxt.getColumnVector(1);
	}

	
} //namespace LibClassicalCrypto
