/***************************************************************************
 *   Copyright (C) 2007 by apterium   *
 *   apterium@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __LIB_NUMBER_THEORY_ERATOSTHENE_H
#define __LIB_NUMBER_THEORY_ERATOSTHENE_H

#include <vector>
#include <gmpxx.h>

namespace libNumberTheory
{

	typedef std::vector<bool>::size_type size_type;
	/**
	 * @short Eratosthene sieve for generating the list of prime numbers.
	 * The Eratosthene sieve. This class is intended to create the list of
	 * prime numbers less or equal than a given bound. Such a list will
	 * necessary for all smooth-number based algorithms (mainly in
	 * cryptanalysis, but also in elliptic curve point counting etc.).
	 * This algorithm current implementation does not use any tricks to
	 * reduce memory usage and improve speed (for example, remembering only the
	 * differences between numbers in order to gain memory, not considering
	 * multiple of 2 at the beginning, etc.), especially because this
	 * algorithm is basically needed to be ran only once.
	 * Note that the current implementation is limited by the max size of
	 * the size_type. On a 64 bits processors, that is 2^64-1. Who would
	 * like to generate such a prime list anyway ?
	 * 
	 * @author apterium
	 */
	class Eratosthene
	{
	public:
		/**
		 * Constructs a new Eratosthene engine instance, given the bound
		 * and a file name.
		 * 
		 * @param bound the bound until which compute primes. MUST be
		 * less than or equal to the max value of
		 * std::vector<boolean>::size_type, that often is the max value
		 * of an unsigned int.
		 */ 
		Eratosthene(const mpz_class& b);
		
		
		~Eratosthene();
		
		/**
		 * Tells whether every primes between 2 and bound have been got
		 * using getNextPrime(). 
		 * 
		 * @return true if and only if some prime number remains which
		 * was not returned yet.
		 */
		bool isOver();
		
		/**
		 * Incrementally gets the prime numbers between 2 and bound,
		 * beginning with 2. Don't hope to get them back if you miss
		 * one.
		 *
		 * @return the current 
		 */
		mpz_class getNextPrime();

	private:
		/**
		 * The maximal bound until which we want to computes the prime
		 * numbers.
		 * 
		 */
		size_type bound;
		
		/**
		 * The maximal index before which we have to sieve. That is,
		 * ceil(sqrt(bound)).
		 *
		 */
		size_type maxIndex;
		
		/**
		 * After the computations, vect->at(i) = 1 iff i is prime.
		 * Indexes begin at 0, 0 corresponds to 0. The computation
		 * begins with index 2.
		 * 
		 */
		std::vector<bool>* vect;

		/**
		 * The current index of the current found prime number. Note
		 * that it correspond to the index in the vector.
		 * 
		 */
		size_type current;
	};

} //namespace libNumberTheory

#endif //__LIB_NUMBER_THEORY_ERATOSTHENE_H
