/***************************************************************************
 *   Copyright (C) 2007 by apterium   *
 *   apterium@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __LIB_ALGEBRA_DENSE_MATRIX_H
#define __LIB_ALGEBRA_DENSE_MATRIX_H

#include <iostream>
#include <vector>
#include <gmpxx.h>
#include "DimensionMismatchException.h"
#include "NonInvertibleMatrixException.h"

namespace libalgebra
{
	/**
	 * @short Basic prototype for dense matrices over a Ring.
	 *
	 * This class provides template implementation and operators on matrices
	 * of Ring Elements.
	 * 
	 * T is the Ring Element type. It must implement the RingElement interface.
	 * R is the Ring type, that is this matrix is a matrix over R. T and R
	 * must correspond.
	 * @todo protect the access to the underlying matrix with exceptions
	 * thrown by the public member accession functions if the index is out
	 * of range. (for example, getRowVector ).
	 * @author apterium
	 */
	template<class T, class R> 
	class DenseMatrix
	{
	public:
		typedef typename std::vector<std::vector<T> >::size_type size_type;
		/**
		 * Constructs a new matrix from its representation into a vector
		 * of vector. Note that it does not verify whether the lengths
		 * of the inner vectors are identical. If it's not the case, you
		 * agree that you expect index out of range errors, etc. 
		 * If the provided vector is empty, the program will abort in
		 * the best case. You have been warned...
		 *
		 * @param content a reference to the content of the matrix.
		 * @param r a constant pointer to the Ring over which this
		 * matrix is defined.
		 */
		DenseMatrix(const std::vector<std::vector<T> >& content, const R* r);

		/**
		 * Copy constructor. Copies the content of the parameter into a
		 * new matrix. The former one can thus be deleted safely.
		 *
		 * @param matrix a constant reference to the matrix which will
		 * be copied.
		 */
		DenseMatrix(const DenseMatrix<T, R> &matrix);
		
		/**
		 * Constructs a new matrix, intialised to the default value of
		 * T, given its dimensions.
		 *
		 * @param rows a constant reference to the number of rows.
		 * @param columns a constant reference to the number of columns.
		 * @param r a constant pointer to the ring which this matrix is
		 * defined over.
		 */
		DenseMatrix(const mpz_class& rows, const mpz_class& columns, const R* r);

		/**
		 * Destructor.
		 *
		 */
		~DenseMatrix();

		/**
		 * Returns the identity matrix of size dimension.
		 * 
		 * @param dimension a constant reference to the dimension of the matrix.
		 * @param r a constant pointer to the ring this matrix will be
		 * defined on.
		 * @return the identity matrix
		 */
		static DenseMatrix<T, R> identity(const mpz_class& dimension,
						  const R* r);

		/**
		 * Constructs a new unicolumn dense matrix, filled with the
		 * content of a given vector.
		 * 
		 * @param vec a constant reference to the vector which is to be represented as a column
		 * vector.
		 * @param r a constant pointer to the ring this vector is
		 * defined on.
		 */
		static DenseMatrix<T, R> columnVector(const std::vector<T>& vec,
						       const R* r);
		
		/**
		 * Constructs a new unirow dense matrix, given the content of a
		 * vector. Same as column vector, but for rows.
		 *
		 * @see columnVector
		 */
		static DenseMatrix<T, R> rowVector(const std::vector<T>& vec,
						   const R* r);

		/**
		 * Gets the rown number.
		 * 
		 */
		mpz_class getRowDimension() {
			return mpz_class(this->rowsDimension);
		};
		
		mpz_class getColumnDimension() {
			return mpz_class(this->columnsDimension);
		};

		/**
		 * Gets the (row, column) coefficient of the matrix. Note that
		 * the row/column numbers have to lie between 1 and the
		 * respective dimension. 
		 *
		 * @todo add a specific exception if the index is out of bounds ?
		 * @param row the row number where to take the coefficient.
		 * @param column the column number where to take the coefficient.
		 * @return a constant reference to the element, of type T.
		 */
		const T& getElement(const mpz_class& row, const mpz_class& column) const;
		
		/**
		 * Gets a transpose matrix of this one. Note that the values are
		 * copied.
		 *
		 * @todo todo ! Carefully think about how it should be done,
		 * given the representation... Maybe use a double matrix instead
		 * of a vector of vectors ? But would be very nice indeed !!
		 */
		DenseMatrix<T, R> getTranspose() const;
		
		/**
		 * Transposes this matrix. Note that this matrix is transformed
		 * to its transpose matrix, if you don't want to modify this use
		 * getTranspose() instead. 
		 * 
		 */
		void transpose();
		
		/**
		 * Constructs the minor matrix, obtained by removing one row and
		 * one column to this matrix.
		 *
		 * @param row the row to be removed.
		 * @param column the column which shall be removed.
		 * @return the minor matrix.
		 */
		DenseMatrix<T, R> cofactorSubMatrix(const mpz_class& row, const mpz_class& column) const;
		
		/**
		 * Compute the addition of a matrix on T to this one.
		 *
		 * @param right a constant reference to the matrix which is
		 * about to be added.
		 * @return the addition of both matrices.
		 * @throw DimensionMismatchException if the dimensions are not
		 * the same.
		 */
		DenseMatrix<T, R> operator+(const DenseMatrix<T, R> &right) const
			throw(DimensionMismatchException);
		
		
		/**
		 * Computes the subtraction of a matrix on T to this one.
		 *
		 * @param right a constant reference to the matrix which is
		 * about to be subtracted.
		 * @return the subtraction of both matrices, that is this minus
		 * right
		 * @throw DimensionMismatchException if the dimensions are not
		 * the same.
		 */
		DenseMatrix<T, R> operator-(const DenseMatrix<T, R> &right) const
			throw(DimensionMismatchException);


		/**
		 * Computes the product of this matrix by the matrix T.
		 *
		 * @param right a constant reference to the matrix which is
		 * about to be right-multiplied to this one.
		 * @return the product of both matrices, that is this times
		 * right
		 * @throw DimensionMismatchException if the dimensions are not
		 * conveninent, that is, if this has not as many columns as
		 * right has rows.
		 */
		DenseMatrix<T, R> operator*(const DenseMatrix<T, R> &right) const
			throw(DimensionMismatchException);

		/**
		 * Computes the inverse of this matrix. The matrix should be
		 * invertible, else a NonInvertibleMatrixException is thrown.
		 *
		 * @return the inverse of this matrix.
		 * @throw DimensionMismatchException if the matrix is not a
		 * square matrix
		 * @throw NonInvertibleMatrixException if the matrix is not invertible.
		 */
		DenseMatrix<T, R> inverse() const 
			throw(DimensionMismatchException, NonInvertibleMatrixException);
		
		/**
		 * Gets the row th row vector. That is, (M(row,j)) with
		 * j=1...columnsDimension.
		 * 
		 * @param row a constant reference to the row number.
		 * @return a std::vector<T> containing the row vector
		 */
		std::vector<T> getRowVector(const mpz_class& row) const;

		/**
		 * Gets the column th column vector. That is, the vector (M(i,
		 * column) with i = 1...rowsDimension.
		 * 
		 * @param column a constant reference to the column index.
		 * @return a std::vector<T> containing the desired column vector.
		 * 
		 */
		std::vector<T> getColumnVector(const mpz_class& column) const;
		
		/**
		 * Computes the determinant of this matrix. It must naturally be
		 * a square matrix. 
		 *
		 * @return the value of the determinant (Ring Element T).
		 * @throw DimensionMismatchException if the matrix is not
		 * square.
		 */
		T determinant() const throw(DimensionMismatchException);
		
		/**
		 * Pretty printing of a matrix on a stream.
		 *
		 * @param out the output stream.
		 * @param matrix the matrix which is going to be sent on the
		 * stream.
		 * @return the resulting outputstream.
		 */ 
		template<class U, class RR> 
		friend std::ostream& operator<<(std::ostream& out, const DenseMatrix<U, RR>& matrix);

		/**
		 * Basic operator, multiplying a matrix by an element of the
		 * ring results in the matrix where all coefficients are
		 * multiplied by this one.
		 *
		 * @param coeff a constant reference to the multiplying coefficient
		 * @param matrix a constant reference to the matrix to be
		 * multiplied by.
		 * @return the resulting matrix.
		 */
		template<class U, class RR>
		friend DenseMatrix<U, RR> operator*(const U& coeff, 
						    const DenseMatrix<U, RR>& matrix);


		
	private:
		//CONVENTION : 
		//the matrix is represented by the vector of its rows. That is,
		//matr->at(i) is the ith row vector of the matrix.
		std::vector<std::vector<T> >* matr;
		
		//facilitate the wide usage of loops in the different operators,
		//etc. 
		//To be maintained accurate at the creation of the
		//matrix, for example when multiplying two matrices
		size_type rowsDimension;
		size_type columnsDimension;

		const R* ring;
		
		/**
		 * Multiplies this by right, on the right, using the standard
		 * (O(n^3) ) algorithm. Note that the matrix MUST be compatible
		 * for the product, that is their dimensions are OK.
		 *
		 * @param right a constant reference to the matrix we multipky with.
		 * @return (*this) * right
		 */ 
		DenseMatrix<T, R> standardMultiplication(const DenseMatrix<T, R>& right) const;

		/**
		 * Computes the cofactor matrix of this, on row, column. It is
		 * copied so using this method is memory hungry if used recursively.
		 *
		 * @param row the row to be deleted.
		 * @param column the column to be deleted.
		 * @return the cofactor of this, that is (-1)^(i+j) * M where M is
		 * the matrix (*this) whithout the row i and the column j.
		 */
		DenseMatrix<T, R> cofactSubMatr(const size_type& row, const size_type& column) const;

		/**
		 * Calculates the cofactors matrix (comatrix) using computation
		 * of determinant of all submatrices. Note this implies that the
		 * matrix is a square matrix, as the public methods ensure.
		 * 
		 *
		 */
		DenseMatrix<T, R> cofactorsMatrix() const;
		
		/**
		 * Calculates the determinant, using the Laplace row/column
		 * expansion, in a naive way. That is, the expansion is
		 * calculated expanding the first line, then recursively.
		 * 
		 * Note that this method is provided almost only for the case
		 * where matrix Gauss reduction is not possible, for example on
		 * a non integral domain Ring (cf the Hill matrix inversion problem) : there exists
		 * matrices on Z/26Z which are invertible even if every element
		 * of it is a non invertible element of Z/26Z. 
		 * May be fasten first with a hardcoded formula for the determinant for small
		 * dimension matrices ( n <= 10 for example ?) if really to
		 * slow, then by stopping computation if we fall on a zero
		 * coefficient.
		 * May be used after an attempt of Gauss reduction, having
		 * obtained a maximal possible reduction ?
		 * May be parallelised (multithreaded) !!
		 * May be improved using pointers and sharing the memory, or
		 * using a different implementation of the matrices (we could
		 * recursively give the deleted columns and lines by using a list, and
		 * iterating on the list and do not copy the whole values as minMatr does).
		 *
		 * @return the determinant of this matrix.
		 */
		T naiveLaplaceExpansion() const;

		/**
		 * @see getRowVector
		 * Identical, but not protected against bad index range.
		 * 
		 */
		std::vector<T> getRowVect(const size_type& row) const;

		/**
		 * Same as getRowVect but for column vector.
		 * @see getRowVect
		 *
		 */
		std::vector<T> getColumnVect(const size_type& column) const;
		
		/**
		 * Default constructor, that is, construct a new, empty,
		 * DenseMatrix.
		 * It is private because it shall not be used without using initialize.
		 *
		 * @param r a constant pointer to the ring to be used.
		 * @todo replace the couple DenseMatrix() + initialize() with a
		 * private static function which would avoid initialize(), whose
		 * role may be directly achieved with normal vector construction.
		 */ 
		DenseMatrix(const R* r)
		{
			this->ring = r;
		};
		
		/**
		 * Initialize this matrix provided its number of rows and of
		 * columns. It is filled with the default value of T.
		 *
		 * @param rows the number of rows.
		 * @param columns the number of columns.
		 * @param r the ring
		 */
		void initialize(const size_type& rows, 
				const size_type& columns);

		/**
		 * Gets the element, private version with size_type.
		 * 
		 * @see getElement(const mpz_class&, const mpz_class&);
		 */
		inline const T& getElt(const size_type& row, const size_type& column) const
		{
			return (this->matr)->at(row).at(column);
		};
		
		
	};
	


} //namespace libalgebra


#include "DenseMatrix.tpp"

#endif //__LIB_ALGEBRA_DENSE_MATRIX_H
