/***************************************************************************
 *   Copyright (C) 2007 by apterium   *
 *   apterium@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

namespace libalgebra
{
	template<class T, class R> 
	DenseMatrix<T, R>::DenseMatrix(const std::vector<std::vector<T> > &content,
				    const R* r)
	{
		this->ring = r;
		this->matr = new std::vector<std::vector<T> >(content);
		this->rowsDimension = (this->matr)->size();
		this->columnsDimension = (this->matr)->at(0).size();
	}
	
	template<class T, class R>
	DenseMatrix<T, R>::DenseMatrix(const DenseMatrix<T, R> &matrix)
	{
		this->ring = matrix.ring;
		this->matr = new std::vector<std::vector<T> >(*(matrix.matr));
		this->rowsDimension = matrix.rowsDimension;
		this->columnsDimension = matrix.columnsDimension;
	}

	template<class T, class R>
	DenseMatrix<T, R>::DenseMatrix(const mpz_class& rows, 
				       const mpz_class& columns,
				       const R* r)
	{
		size_type lin = (size_type) rows.get_ui();
		size_type col = (size_type) columns.get_ui();
		
		#ifdef DEBUG
		std::cout << std::endl << "Line " << __LINE__ << " of " << __FILE__  << std::endl;
		std::cout << "Rows mpz_class : " << rows << " Rows size_type : " << lin << std::endl;
		std::cout << "Columns mpz_class : " << columns << " Columns size_type : " << col << std::endl; 
                #endif //DEBUG
		this->ring = r;
		this->initialize(lin, col);
	}

	template<class T, class R>
	DenseMatrix<T, R>::~DenseMatrix()
	{
		delete this->matr;
	}

	template<class T, class R>
	DenseMatrix<T, R> DenseMatrix<T, R>::identity(const mpz_class& dimension, 
						      const R* r)
	{
		DenseMatrix<T, R> ret = DenseMatrix(dimension, dimension, r);
		for (size_type i = 0 ; i < ret.rowsDimension ; i++) {
			(ret.matr)->at(i).at(i) = T(1, r);
		}
		return ret;
	}
	
	template<class T, class R>
	DenseMatrix<T, R> DenseMatrix<T, R>::columnVector(const std::vector<T>& vec,
							  const R* r)
	{
		DenseMatrix<T, R> ret = DenseMatrix(mpz_class(vec.size()), 1, r);
		for (size_type i = 0 ; i < vec.size() ; i++) {
			(ret.matr)->at(i).at(0) = vec.at(i);
		}
		return ret;
	}

	template<class T, class R>
	DenseMatrix<T, R> DenseMatrix<T, R>::rowVector(const std::vector<T>& vec,
							  const R* r)
	{
		DenseMatrix<T, R> ret = DenseMatrix(1, mpz_class(vec.size()), r);
		for (size_type i = 0 ; i < vec.size() ; i++) {
			(ret.matr)->at(0).at(1) = vec.at(i);
		}
		return ret;
	}


	
	template<class T, class R>
	const T& DenseMatrix<T, R>::getElement(const mpz_class& row, const mpz_class& column) const
	{
		size_type lin = (size_type) row.get_ui() - 1;
		size_type col = (size_type) column.get_ui() - 1;
		
#ifdef DEBUG
		std::cout << std::endl << "Line " << __LINE__ << " of " << __FILE__  << std::endl;
		std::cout << "Rows mpz_class : " << row << " Row size_type : " << lin << std::endl;
		std::cout << "Columns mpz_class : " << column << " Column size_type : " << col << std::endl; 
#endif //DEBUG
		return this->getElt(lin, col);
	}

	template<class T, class R>
	DenseMatrix<T, R> DenseMatrix<T, R>::getTranspose() const
	{
		DenseMatrix<T, R> ret = DenseMatrix(this->ring);
		ret.initialize(this->columnsDimension, 
			       this->rowsDimension);
		for (size_type i = 0 ; i < this->rowsDimension ; i++) {
			for (size_type j = 0 ; j < this->columnsDimension ; j++) {
				(ret.matr)->at(j).at(i) = this->getElt(i, j);
			}
		}
		return ret;
	}
	
	template<class T, class R>
	DenseMatrix<T, R> DenseMatrix<T, R>::cofactorSubMatrix(const mpz_class& row, const mpz_class& column) const
	{
		size_type lin = (size_type) row.get_ui() - 1;
		size_type col = (size_type) column.get_ui() - 1;
		
		#ifdef DEBUG
		std::cout << std::endl << "Line " << __LINE__ << " of " << __FILE__  << std::endl;
		std::cout << "Rows mpz_class : " << row << " Row size_type : " << lin << std::endl;
		std::cout << "Columns mpz_class : " << column << " Column size_type : " << col << std::endl; 
                #endif //DEBUG
		return this->cofactSubMatr(lin, col);
	}
	
	template<class T, class R>
	DenseMatrix<T, R> DenseMatrix<T, R>::operator+(const DenseMatrix<T, R>& right) const 
		throw(DimensionMismatchException)
	{
		if ( (this->rowsDimension != right.rowsDimension) ||
		     (this->columnsDimension != right.columnsDimension) ) {
			throw DimensionMismatchException();
		}
		DenseMatrix<T, R> ret(*this);
		for (size_type i = 0 ; i < this->rowsDimension ; i++) {
			for (size_type j = 0 ; j < this->columnsDimension ; j++) {
				(ret.matr)->at(i).at(j) += (right.matr)->at(i).at(j);
			}
		}
		return ret;
	}
	
	template<class T, class R>
	DenseMatrix<T, R> DenseMatrix<T, R>::operator-(const DenseMatrix<T, R>& right) const
		throw(DimensionMismatchException)
	{
		if ( (this->rowsDimension != right.rowsDimension) ||
		     (this->columnsDimension != right.columnsDimension) ) {
			throw DimensionMismatchException();
		}
		DenseMatrix<T, R> ret(*this);
		for (size_type i = 0 ; i < this->rowsDimension ; i++) {
			for (size_type j = 0 ; j < this->columnsDimension ; j++) {
				(ret.matr)->at(i).at(j) -= (right.matr)->at(i).at(j);
			}
		}
		return ret;
	}
	
	template<class T, class R>
	DenseMatrix<T, R> DenseMatrix<T, R>::operator*(const DenseMatrix<T, R>& right) const
		throw(DimensionMismatchException)
	{
		if (this->columnsDimension != right.rowsDimension) {
			throw DimensionMismatchException();
		}
		return this->standardMultiplication(right);
	}
	
	template<class T, class R> 
	DenseMatrix<T, R> DenseMatrix<T, R>::inverse() const
		throw(DimensionMismatchException, NonInvertibleMatrixException)
	{

		if (this->rowsDimension != this->columnsDimension) {
			throw DimensionMismatchException();
		}
		T det = this->determinant();
		if (!(det.isInvertible()) ) {
			throw NonInvertibleMatrixException();
		}
		if (this->rowsDimension == 1) {
			//just one element...
			return columnVector(std::vector<T>(1, det.inverse()), this->ring);
		}
		return det.inverse() * ((this->cofactorsMatrix()).getTranspose());
	}

	template<class T, class R>
	T DenseMatrix<T, R>::determinant() const throw(DimensionMismatchException)
	{
		if (this->rowsDimension != this->columnsDimension) {
			throw DimensionMismatchException();
		}
		return this->naiveLaplaceExpansion();
	}
	
	template<class T, class R>
	std::vector<T> DenseMatrix<T, R>::getRowVector(const mpz_class& row) const
	{
		size_type lin = (size_type) row.get_ui() - 1;
		#ifdef DEBUG
		std::cout << std::endl << "Line " << __LINE__ << " of " << __FILE__  << std::endl;
		std::cout << "Row mpz_class : " << row << " Row size_type : " << lin << std::endl;
                #endif //DEBUG
		
		return this->getRowVect(lin);
	}
	
	template<class T, class R>
	std::vector<T> DenseMatrix<T, R>::getColumnVector(const mpz_class& column) const
	{
		size_type col = (size_type) column.get_ui() - 1;
		#ifdef DEBUG
		std::cout << std::endl << "Line " << __LINE__ << " of " << __FILE__  << std::endl;
		std::cout << "Column mpz_class : " << column << " Column size_type : " << col << std::endl;
		#endif
		return this->getColumnVect(col);
	}
	
	template<class U, class RR> 
	std::ostream& operator<<(std::ostream& out, 
				 const DenseMatrix<U, RR>& matrix)
	{
		for (typename std::vector<std::vector<U> >::size_type i = 0 ; 
		     i < matrix.rowsDimension ; i++) {
			out << std::endl;
			for (typename std::vector<U>::size_type j = 0 ; 
			     j < matrix.columnsDimension ; j++) {
				out << matrix.getElt(i,j) << " ";
			}
		}
		return out;
	}

	template<class U, class RR>
	DenseMatrix<U, RR> operator*(const U& coeff, const DenseMatrix<U, RR>& matrix)
	{
		DenseMatrix<U, RR> ret = DenseMatrix<U, RR>(matrix.ring);
		ret.initialize(matrix.rowsDimension, matrix.columnsDimension);
		for (typename std::vector<std::vector<U> >::size_type i = 0 ; 
		     i < matrix.rowsDimension ; i++) {
			for (typename std::vector<U>::size_type j = 0 ; 
			     j < matrix.columnsDimension ; j++) {
				(ret.matr)->at(i).at(j) = coeff * matrix.getElt(i, j);
			}
		}
		return ret;
	}

	template<class T, class R>
	DenseMatrix<T, R> DenseMatrix<T, R>::standardMultiplication(const DenseMatrix<T, R>& right) const
	{
		DenseMatrix<T, R> ret(this->rowsDimension, right.columnsDimension, this->ring);
		for (size_type i = 0 ; i < this->rowsDimension ; i++) {
			for (size_type j = 0 ; j < right.columnsDimension ; j++) {
				//default allocation, that is 0 and the ring !
				T cij(mpz_class(0), this->ring);
				for (size_type k = 0 ; k < this->columnsDimension ; k++) {
					cij += (this->matr)->at(i).at(k) * (right.matr)->at(k).at(j);
				}
				(ret.matr)->at(i).at(j) = cij;
			}
		}
		return ret;
	}

	template<class T, class R>
	void DenseMatrix<T, R>::initialize(const size_type &rows, const size_type &columns)
	{
		//first create the rows : length columns !
		std::vector<T> row(columns, T(mpz_class(0), this->ring));
		//then copy it column times into a new vector
		this->matr = new std::vector<std::vector<T> >(rows, row);
		this->rowsDimension = rows;
		this->columnsDimension = columns;
	}
	
	template<class T, class R> 
	DenseMatrix<T, R> DenseMatrix<T, R>::cofactSubMatr(const size_type& row, const size_type& column) const
	{
		DenseMatrix<T, R> ret = DenseMatrix(this->ring); 
		ret.initialize(this->rowsDimension - 1, this->columnsDimension - 1);

		//the multiplicative coefficient
		T coeff(this->ring) ;
		if ( (row + column) % 2 == 0) {
			coeff = T(1, this->ring);
		} else {
			coeff = T(-1, this->ring);
		}
		//then insert coefficients into the new matrix, avoiding the row and the column
		for (size_type i = 0 ; i < row ; i ++) {
			for (size_type j = 0 ; j < column ; j++) {
				(ret.matr)->at(i).at(j) = coeff * (this->getElt(i, j));
			}
			for (size_type j = column + 1 ; j < this->columnsDimension ; j++) {
				(ret.matr)->at(i).at(j - 1) = coeff * (this->getElt(i, j));
			}
		}
		for (size_type i = row + 1 ; i < this->rowsDimension ; i++) {
			for (size_type j = 0 ; j < column ; j++) {
				(ret.matr)->at(i - 1).at(j) = coeff * (this->getElt(i, j));
			}
			for (size_type j = column + 1 ; j < this->columnsDimension ; j++) {
				(ret.matr)->at(i - 1).at(j - 1) = coeff * (this->getElt(i, j));
			}
		}
		return ret;
	}

	template<class T, class R>
	DenseMatrix<T, R> DenseMatrix<T, R>::cofactorsMatrix() const
	{

		DenseMatrix<T, R> ret = DenseMatrix(this->ring);
		ret.initialize(this->rowsDimension, this->columnsDimension);
		
		for (size_type i = 0 ; i < this->rowsDimension ; i++) {
			for (size_type j = 0 ; j < this->columnsDimension ; j++) {
				DenseMatrix<T, R> currCofactor = this->cofactSubMatr(i, j);
				(ret.matr)->at(i).at(j) = currCofactor.determinant();
			}
		}
		return ret;
	}
	
	template<class T, class R>
	std::vector<T> DenseMatrix<T, R>::getRowVect(const size_type& row) const
	{
		std::vector<T> ret;
		ret.reserve(row);
		for (size_type i = 0 ; i < this->columnsDimension - 1 ; i++) {
			ret.push_back(this->matr->at(row).at(i) );
		}
		return ret;
	}
	
	template<class T, class R>
	std::vector<T> DenseMatrix<T, R>::getColumnVect(const size_type& column) const
	{
		std::vector<T> ret;
		ret.reserve(this->rowsDimension);
		for (size_type i = 0 ; i < this->rowsDimension ; i++) {
			ret.push_back(this->matr->at(i).at(column));
		}
		return ret;
	}

	template<class T, class R>
	T DenseMatrix<T, R>::naiveLaplaceExpansion() const
	{
		//development following the first line
		T ret = T(mpz_class(0), this->ring);
		if (this->columnsDimension > 1) {
			for (size_type i = 0 ; i < this->rowsDimension ; i++) {
				ret += (this->matr)->at(0).at(i) * (this->cofactSubMatr(0,i)).determinant();
			}	       
		} 
		else {  //the matrix has only one element
			//might better be in the determinant function directly
			ret += this->matr->at(0).at(0);
		}
		return ret;
	}

	
} //namespace libalgebra
