/***************************************************************************
 *   Copyright (C) 2007 by apterium   *
 *   apterium@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __LIB_ALGEBRA_DENSE_POLYNOMIAL_H
#define __LIB_ALGEBRA_DENSE_POLYNOMIAL_H

#include <iostream>
#include <vector>
#include <gmpxx.h>
#include "RingElement.h"
#include "NonInvertibleElementException.h"
#include "RingMismatchException.h"

namespace libalgebra
{
	/**
	 * @short Basic prototype for dense polynomials over a Ring.
	 *
	 * T is the Ring Element type. Note that it MUST provide at least
	 * basic ring operations (including inversion when possible)  and an
	 * operator == to compare two T values.
	 * R is the basic Ring type. It must obviously correspond to the
	 * RingElement type...
	 * @todo add the modular inversion function. It will use the Euclidean
	 * algorithm which will be in a templatised function somewhere. Don't
	 * forget to throw an exception if both polynoms are not coprime !
	 * @todo add a generic polynomial class, with the following methods : 
	 * -a getUnderlyingRing() to get the Ring the coefficients of this
	 * polynom belongs to.
	 * -a degree() method.
	 * -a coefficient(const mpz_class& ) method.
	 * -an evaluation method, which gives the evaluation of the polynomial
	 * at a given point.
	 *
	 * @author apterium
	 */
	template<class T, class R>
	class DensePolynomial// : public RingElement
	{
	public:
		typedef typename std::vector<T>::size_type size_type;
		
		/**
		 * Constructs a new polynomial, given its coefficients. Will copy
		 * them so the argument must be deleted by the caller.
		 *
		 * @param coefficients an ordered vector containing the
		 * polynomial's coefficients in increasing degree order.
		 * @param r a constant pointer to the ring this polynomial will
		 * be based on.
		 */
		DensePolynomial(const std::vector<T>& coefficients, const R* r);
		
		/**
		 * Copy constructor. That is, copy the internal structure of a
		 * polynomial into the new one.
		 *
		 * @param polynomial the polynomial whose copy shall be created.
		 */
		DensePolynomial(const DensePolynomial<T, R>& polynomial);
		
		/**
		 * Constructs a new, constant equal to value, polynomial.
		 *
		 * @param value a constant reference to the value the polynomial
		 * will take.
		 * @param r a constant pointer to the ring the new polynomial
		 * will be based on.
		 */
		DensePolynomial(const T& value, const R* r);
		
		/**
		 * Destructor.
		 *
		 */
		~DensePolynomial();
		
		/**
		 * Gets the degree of this polynomial. Note that the null polynomial
		 * (identically equal to zero) is reporter to have the degree 0.
		 *
		 * @return the degree in question. It must
		 * be deleted by the caller.
		 */
		inline mpz_class degree() const;
		
		/**
		 * Gets the coefficient of X^i in this polynomial. IMPORTANT NOTE :
		 * i must fit in an unsigned long integer, and be convertible
		 * into the convenient vector<T, R>::size_type.
		 *
		 * @todo this method should be in a parent class, a  generic
		 * polynomial should provide at least a method to get its
		 * coefficients.
		 * @todo it does not seem to be a clean vay to do portable code
		 * : size_type is not necessarily an ui...
		 * @param i a reference to the degree of the requested coefficient.
		 * @return a constant reference to the coefficient, which is of type T.
		 */ 
		inline T coefficient(const mpz_class& i) const;
		
		/**
		 * Evaluates this polynomial at one point. Note it must be the
		 * same type as the polynomial's coefficients.
		 * Uses Horner algorithm in order to be fast and efficient.
		 *
		 * @param value a reference towards the evaluation point.
		 * @return the evaluation value. Must be
		 * deleted by the caller.
		 * @throw RingMismatchException if the elements does not belong
		 * to the same ring.
		 */
		T eval(const T& value) const throw(RingMismatchException);
		
		/**
		 * Return the ring this polynomials' coefficients belong to.
		 * 
		 * @return a constant reference to the Ring in question.
		 */
		inline const R* getUnderlyingRing() const;
		
		/**
		 * Adds a dense polynomial to this one. 
		 *
		 * @return the result. The memory must be freed
		 * by the caller.
		 * @throw RingMismatchException if a ring mismatch is detected
		 * during the computations.
		 */
		DensePolynomial<T, R> operator+(const DensePolynomial<T, R>& right) const
			throw(RingMismatchException);
		
		/**
		 * Substracts a dense polynomial to this one. 
		 *
		 * @return the result. The memory must be freed
		 * by the caller.
		 * @throw RingMismatchException if a ring mismatch is detected
		 * during the computations.
		 */
		DensePolynomial<T, R> operator-(const DensePolynomial<T, R>& right) const
			throw(RingMismatchException);

		/**
		 * Multiply a dense polynomial by this one.
		 *
		 * @return the result. The memory must be freed
		 * by the caller.
		 */
		DensePolynomial<T, R> operator*(const DensePolynomial<T, R>& right) const
			throw(RingMismatchException);
		
		// NOTE ON DIVISION and REMAINDER OPERATORS : 
		// should be sure that head coefficient of the dividend is
		// invertible. So try and see if a NonInvertibleElementException
		// is thrown.
		

		/**
		 * Gets the quotient in the euclidean division of this polynomial
		 * by an other one. The leading coefficient (coefficient of
		 * X^degre in the divisor) of the divisor must be
		 * invertible for the operation to be well-defined.
		 * 
		 * @param divisor a constant reference to the divisor.
		 * @return the quotient of this by divisor.
		 * @throw NonInvertibleElementException<T, R> if the leading
		 * coefficient of the divisor is not invertible.
		 */
		DensePolynomial<T, R> operator/(const DensePolynomial<T, R>& divisor) const
			throw(NonInvertibleElementException<T>,
			      RingMismatchException);
		
		/**
		 * Gets the remainder in the euclidean division of this polynomial
		 * by modulus. The leading coefficient of modulus must be invertible.
		 *
		 * @param modulus the divisor of the euclidean division.
		 * @return the remainder of the division of this by the modulus.
		 * @throw NonInvertibleElementException if the leading
		 * coefficient of the modulus is not invertible.
		 */
		DensePolynomial<T, R> operator%(const DensePolynomial<T, R>& modulus) const
			throw(NonInvertibleElementException<T>,
			      RingMismatchException);
		
		/**
		 * Tells whether this polynomial is invertible. That is, is it a
		 * constant polynomial whose value is invertible in its ring ?
		 * As const, this function copies the reduction of this
		 * polynomial in a new one, so it may suffer of bad performances.
		 *
		 * @todo May be define it non const in the primary API ?
		 * @return true iff this polynomial is an invertible constant .
		 */ 
		inline bool isInvertible() const;
		
		/**
		 * Delete zero entries at the end of the coefficients. That is,
		 * if the leading coefficient of the polynomial is zero, delete it.
		 * This mechanism allows the degree() method to always return a
		 * correct value, except for the null polynomial.
		 *
		 * @remark : normal use should not call this function at any point.
		 */
		void reduce() throw(RingMismatchException);
		 
	private:
		/**
		 * The coefficients of the vector, ordered with increasing
		 * degree.
		 * 
		 */
		std::vector<T> *coeffs;
		
		/**
		 * The ring this polynomial is based on.
		 * 
		 */
		const R* ring;
		
		/**
		 * Naive multiplication between polynomials. Multiplies this
		 * polynomial by an onther one, following the naive algorithm.
		 *
		 * @param right a constant reference to the right polynomial of the
		 * multiplication.
		 * @return the product of left and right.
		 */
		DensePolynomial<T, R> naiveMultiplication(const DensePolynomial<T, R>& right) const;

		/**
		 * Schoolbook division algorithm. Used recursively and memory
		 * hungry. Should be replaced with a better implementation
		 * handling better the underlying vector.
		 * 
		 */
		DensePolynomial<T, R> naiveDivision(const DensePolynomial<T, R>& divisor) const;

		/**
		 * Gets the ith coefficient of the polynomial, that is, the one
		 * of X^i.
		 *
		 * @param i the coefficient number
		 * @return a constant reference to the ith coefficient.
		 */
		inline T getCoeff(const size_type& i) const
		{
			return (this->coeffs)->at(i);
		};

		/**
		 * Equality operator between two DensePolynomials. Note that it
		 * does a reduction of the polynomials. Perhaps it should not be
		 * the case for efficiency ?
		 * 
		 * @param left a constant reference to the left member of the
		 * operator.
		 * @param right a constant reference to the right member.
		 * @return true iff left and right represent the same
		 * polynomial.
		 */ 
		template<class U, class RR>
		friend bool operator==(DensePolynomial<U, RR> left, DensePolynomial<U, RR> right);

		/**
		 * Non equality operator between two DensePolynomials.
		 * 
		 * @param left a constant reference to the left member of the
		 * non-equality.
		 * @param right a constant reference to the right member of the
		 * non-equality.
		 * @return true iff left and right do not represent the same polynomial
		 */
		template<class U, class RR>
		friend bool operator!=(DensePolynomial<U, RR> left, DensePolynomial<U, RR> right)
		{
			return !(left == right);
		};
		
		/**
		 * Pretty printing of the polynomial.
		 *
		 * @todo better handle the variable name for multivariates polynomials.
		 * @param out a reference to the output stream where the object should be sent.
		 * @param polynomial a reference to the DensePolynomial which is sent.
		 * @return a reference to the resulting output stream.
		 */
		template<class U, class RR> 
		friend std::ostream& operator<<(std::ostream& out, const DensePolynomial<U, RR>& polynomial);
	};

	


} //namespace libalgebra

#include "DensePolynomial.tpp"

#endif //__LIB_ALGEBRA_DENSE_POLYNOMIAL_H
