/***************************************************************************
 *   Copyright (C) 2007 by apterium   *
 *   apterium@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

namespace libalgebra
{
	
	template<class T, class R>
	DensePolynomial<T, R>::DensePolynomial(const std::vector<T>& coefficients, 
					       const R* r)
	{
		this->coeffs = new std::vector<T>(coefficients);
		this->ring = r;
	}
	
 	template<class T, class R>
 	DensePolynomial<T, R>::DensePolynomial(const DensePolynomial<T, R>& polynomial)
 	{
 		this->coeffs = new std::vector<T>(*(polynomial.coeffs));
		this->ring = polynomial.ring;
 	}
	
	template<class T, class R>
	DensePolynomial<T, R>::DensePolynomial(const T& value, const R* r)
	{
		this->coeffs = new std::vector<T>(1, value);
		this->ring = r;
	}
	
	template<class T, class R> 
	DensePolynomial<T, R>::~DensePolynomial()
	{
		if (this->coeffs != NULL) delete this->coeffs;
		coeffs = NULL;
	}

	template<class T, class R> mpz_class DensePolynomial<T, R>::degree() const
	{
		return mpz_class((this->coeffs)->size() - 1);
	}
	
	template<class T, class R>
	T DensePolynomial<T, R>::coefficient(const mpz_class& i) const
	{
		if (i > (this->degree()) ) {
			return T(0, this->ring);
		}
		return this->getCoeff( (size_type) (i.get_ui()));
	}
	
	template<class T, class R> 
	T DensePolynomial<T, R>::eval(const T& value) const
		throw(RingMismatchException)
	{
		//make use of Horner algorithm
		typename std::vector<T>::size_type degree = (this->coeffs)->size() - 1;
		T ret = (this->coeffs)->at(degree);
		for (typename std::vector<T>::size_type i = 0 ; i < degree ; i++) {
			ret *= value;
			ret += (this->coeffs)->at(degree - 1 - i);
		}
		
		return ret;
	}
	
	template<class T, class R>
	const R* DensePolynomial<T, R>::getUnderlyingRing() const
	{
		return this->ring;
	}
	
	template<class T, class R>
	DensePolynomial<T, R> DensePolynomial<T, R>::operator+(const DensePolynomial<T, R>& right) const
		throw(RingMismatchException)
	{
		typename std::vector<T>::size_type lgr = (right.coeffs)->size();
		typename std::vector<T>::size_type lg = (this->coeffs)->size();
	
		std::vector<T> retCoeffs;
		if (lgr > lg) {
			retCoeffs = std::vector<T>(*(right.coeffs));
			typename std::vector<T>::size_type i;
			for (i = 0 ; i < lg; i++) {
				retCoeffs.at(i) += (this->coeffs)->at(i);
			}
		} else {
			retCoeffs = std::vector<T>(*(this->coeffs));
			typename std::vector<T>::size_type i;
			for (i = 0 ; i < lgr; i++) {
				retCoeffs.at(i) += (right.coeffs)->at(i);
			}
		}
		DensePolynomial ret(retCoeffs, this->ring);
		ret.reduce();
		return ret;
	}


	template<class T, class R>
	DensePolynomial<T, R> DensePolynomial<T, R>::operator-(const DensePolynomial<T, R>& right) const
		throw(RingMismatchException)
	{
		typename std::vector<T>::size_type lgr = (right.coeffs)->size();
		typename std::vector<T>::size_type lg = (this->coeffs)->size();
		
		std::vector<T> retCoeffs;
		if (lgr > lg) {
			retCoeffs = std::vector<T>(*right.coeffs);
			typename std::vector<T>::size_type i;
			for (i = 0 ; i < lg; i++) {
				//carefully subtract right to this...
				T val =  (this->coeffs)->at(i) - retCoeffs.at(i);
				retCoeffs.at(i) = val;
			}
			for (i = lg ; i < lgr ; i++) {
				T val = -retCoeffs.at(i);
				retCoeffs.at(i) = val;
			}
		} else {
			retCoeffs = std::vector<T>(*this->coeffs);
			typename std::vector<T>::size_type i;
			for (i = 0 ; i < lgr; i++) {
				retCoeffs.at(i) -= (right.coeffs)->at(i);
			}
		}
		DensePolynomial ret(retCoeffs, this->ring);
		ret.reduce();
		return ret;
		
	}
	
	template<class T, class R>
	DensePolynomial<T, R> DensePolynomial<T, R>::operator*(const DensePolynomial<T, R>& right) const
		throw(RingMismatchException)
	{
		return this->naiveMultiplication(right);
	}

	template<class T, class R>
	DensePolynomial<T, R> DensePolynomial<T, R>::operator/(const DensePolynomial<T, R>& divisor) const
		throw(NonInvertibleElementException<T>,
		      RingMismatchException)
	{
		T headCoeff = divisor.coefficient(divisor.degree());
		if ( !(headCoeff.isInvertible()) ) {
			throw NonInvertibleElementException<T>(headCoeff);
		}
		return naiveDivision(divisor);
	}
	
	template<class T, class R>
	DensePolynomial<T, R> DensePolynomial<T, R>::operator%(const DensePolynomial<T, R>& modulus) const
		throw(NonInvertibleElementException<T>,
		      RingMismatchException)
	{
		return (*this) - (*this / modulus) * modulus;
	}
	
	template<class T, class R>
	bool DensePolynomial<T, R>::isInvertible() const
	{
		DensePolynomial<T, R> temp = DensePolynomial<T, R>(*this);
		temp.reduce();
		return ( (temp.coeffs)->size() == 0 && 
			 (temp.coeffs)->at(0).isInvertible() );
	}
	
	template<class U, class RR> 
	std::ostream& operator<<(std::ostream& out, const DensePolynomial<U, RR>& polynomial)
	{
		mpz_class deg = polynomial.degree();
		//avoid a (dirty) X^0
		out << polynomial.coefficient(0);
		if (deg == 0) {
			//nothing else to add...
			return out;
		}
		out << " + ";
		for (mpz_class i = 1 ; i <= deg - 1  ; i++) {
			out << polynomial.coefficient(i) << " * X^" << i << " + ";
		}
		out << polynomial.coefficient(deg) << " * X^" << deg ;
		return out;
	}
	
	template<class T, class R>
	DensePolynomial<T, R> DensePolynomial<T, R>::naiveMultiplication(const DensePolynomial<T, R>& right) const
	{
		T zero(mpz_class(0), this->ring);
		std::vector<T> retCoeffs((this->coeffs)->size() + (right.coeffs)->size() - 1, 
					    zero);
		typename std::vector<T>::size_type i;
		typename std::vector<T>::size_type j;
		for (i = 0 ; i < (this->coeffs)->size() ; i++) {
			for (j = 0 ; j < right.coeffs->size() ; j++) {
				retCoeffs.at(i+j) += (this->coeffs)->at(i) * (right.coeffs)->at(j);
			}
		}
		return DensePolynomial(retCoeffs, this->ring);
	}

	template<class T, class R>
	DensePolynomial<T, R> DensePolynomial<T, R>::naiveDivision(const DensePolynomial<T, R>& divisor) const
	{
		size_type degThis = (this->coeffs)->size();
		size_type degDiv = (divisor.coeffs)->size();
		if (degThis < degDiv  ) {
			return DensePolynomial<T, R>(T(0, this->ring), this->ring);
		}
		else if (degThis == degDiv) {
			return DensePolynomial<T, R>(this->getCoeff(degThis-1) / divisor.getCoeff(degDiv - 1), this->ring );
		}
		else {
			//very dirty use of the memory, as the memory is
			//allocated for the DENSE result and should be directly
			//used to store the coefficients as they are calculated.
			DensePolynomial<T, R> leadingTerm = 
				DensePolynomial<T, R>(std::vector<T>(degThis - degDiv + 1, 
								     T(mpz_class(0), this->ring) ),
						      this->ring);
			(leadingTerm.coeffs)->at(degThis - degDiv) = 
				this->getCoeff(degThis - 1) / divisor.getCoeff(degDiv - 1);
			return leadingTerm + ((*this) - leadingTerm * divisor).naiveDivision(divisor);
		}
	}
	
	template<class U, class RR>
	bool operator==(DensePolynomial<U, RR> left, DensePolynomial<U, RR> right)
	{
		left.reduce();
		right.reduce();
		return (*(left.coeffs) == *(right.coeffs));
	}
	
	
	template<class T, class R> 
	void DensePolynomial<T, R>::reduce()
		throw(RingMismatchException)
	{
		T zero(mpz_class(0), this->ring);
		typename std::vector<T>::size_type s = (this->coeffs)->size() -1;
		while ( (s > 0) && ((this->coeffs)->at(s) == zero) ) {
			(this->coeffs)->pop_back();
			s--;
		}
	}


	
} //namespace libalgebra
