/***************************************************************************
 *   Copyright (C) 2007 by apterium   *
 *   apterium@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef __LIB_ALGEBRA_GENERIC_ALGORITHMS_H
#define __LIB_ALGEBRA_GENERIC_ALGORITHMS_H

#include "NonCoprimeElementsException.h"

namespace libalgebra
{
	/**
	 * @short Generic algorithms template methods. For example, Euclidean
	 * algorithm for Euclidean Rings.
	 * 
	 * Note that T must be a RingElement and thus implement the
	 * RingElement.h API. Its Ring must also be euclidean, else the
	 * behaviour of the algorithms is undefined. 
	 * But of course the user will notice it at once when instanciating this
	 * class...
	 * 
	 * @author apterium
	 */
	template<class T>
	class GenericAlgorithms
	{
	public:
		/** 
		 * Computes the greatest common divisor between t1 and t2.
		 * 
		 * @param t1 a constant reference to the first ring element.
		 * @param t2 a constant reference to the second element.
		 * @return the gcd of t1, t2.
		 */
		inline static T gcd(const T& t1, const T& t2);		

		/**
		 * Computes a Bezout equation for value and modulus. That
		 * is, Ring Elements coefficients x and b and the gcd d such that x * value +
		 * b * modulus = d.
		 * Don't use this method when better, faster methods are
		 * available, for example modular inversion of integers may be
		 * done with GMP.
		 *
		 * @param value a constant reference to the value which is to be
		 * inverted.
		 * @param modulus a constant reference to the modulus modulo
		 * which we invert.
		 * @return an array of three T, [d, x, y] where 
		 * x * value + b * modulus = d and d = gcd(value, modulus). It
		 * must be deleted manually, with delete[].
		 */
		inline static std::vector<T> BezoutCoefficients(const T& value, const T& modulus);
		
		/**
		 * Computes the modular inverse of value modulo modulus. They
		 * should be coprime, that is having an invertible gcd in its
		 * ring, else an exception is thrown.
		 * 
		 * @param value a constant reference to the value which is to be
		 * inverted.
		 * @param modulus a constant reference to the modulus modulo
		 * which we invert.
		 * @return x such that value * x = 1 mod modulus.
		 * @throw NonCoprimeElementsException if value and modulus are
		 * not coprime.
		 */
		static T ModularInversion(const T& value, const T& modulus)
			throw(NonCoprimeElementsException);
		
	private:
		/**
		 * Euclidean algorithm, internal recursive version. The ring
		 * which t1 and t2 belong to is supposed to be Euclidean. 
		 * Computes the gcd between t1 and t2.
		 * 
		 * @see gcd
		 */
		static T recursiveEuclideanAlgorithm(const T& t1, const T& t2);
		
		/**
		 * Extended Euclidean algorithm, recursive version.
		 * In the code, the array returned is [d, x, y] where 
		 * x * value + b * modulus = d = gcd(value, modulus).
		 * 
		 */
		static std::vector<T> recursiveExtendedEuclideanAlgorithm(const T& value, 
								const T& modulus);
	};
	

} //namespace libalgebra

#include "GenericAlgorithms.tpp"

#endif //__LIB_ALGEBRA_GENERIC_ALGORITHMS_H
