/***************************************************************************
 *   Copyright (C) 2007 by apterium   *
 *   apterium@gmail.com   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef __LIB_ALGEBRA_RING_Z_N_ELEMENT_H
#define __LIB_ALGEBRA_RING_Z_N_ELEMENT_H

#include <gmpxx.h>
#include <iostream>
#include "RingZn.h"
#include "RingElement.h"
#include "RingMismatchException.h"
#include "NonInvertibleElementException.h"


namespace libalgebra
{
  /**
   * @short Z over nZ Ring Elements.
   *
   * Z/nZ ring elements. Note that every ring operation (that is +, -, *, /
   * etc.) must be done with a RingZnElement which belongs to the same RingZn,
   * else a RingMismatchException will be thrown. Note that the same RingZn
   * really means the same , that is the same object and not an object with the
   * same values.
   *
   * @author apterium
   */
  class RingZnElement : public RingElement
  {
  public:
	  /**
	   * Constructs a new Z/nZ element, given the RingZn it belongs to.
	   * It is equal to zero by default.
	   *
	   * @param r a pointer to the value of the RingZn.
	   */
	  RingZnElement(const RingZn* r);

	  /**
	   * Constructs a new Z/nZ element from an integer.
	   * 
	   * @param k any integer corresponding to the same element modulo n. Its
	   * value is copied.
	   * @param *r a pointer to the corresponding RingZn. The pointer will be
	   * shared by this RingZnElement.
	   */
	  RingZnElement(const mpz_class &k, const RingZn* r);
	  
	  /**
	   * Copy constructor.
	   *
	   * @param copy a reference to the object which is to be copied.
	   */
	  RingZnElement(const RingZnElement &copy);
	  
	  /**
	   * Destructor. It does not destroy the provided ring as it is shared and was
	   * constructed before by someone else.
	   *
	   */
	  ~RingZnElement();
	  
	  /**
	   * Gets the value of this RingZnElement. That is, get the integer between 0
	   * and n-1 corresponding to this element.
	   *
	   * Note that if you modify the returned value it will modify the
	   * RingZnElement itself. If you want to use it elsewhere, be sure it will be
	   * copied ! 
	   *
	   * @return a constant reference to the mpz_class representation of
	   * this number.
	   */
	  const mpz_class& getValue() const
	  {
			  mpz_class &ret = *(this->value);
			  return ret;
	  };

	  inline const RingZn* getRing() const
	  {
		  return this->ring;
	  }
	  
	  /**
	   * Adds two integers modulo n.
	   *
	   * @param right the right Z/nZ element.
	   * @return the value of the addition of both RingZnElement.
	   * @throw RingMismatchException if both elements do not belong to the
	   * same RingZn.
	   */
	  RingZnElement operator+(const RingZnElement& right) const
		  throw(RingMismatchException);

	  /**
	   * Adds a RingZnElement to this one.
	   *
	   * @param right a constant reference to the RingZnElement which is to
	   * be added to this one.
	   */
	  void operator+=(const RingZnElement& right)
		  throw(RingMismatchException);
	  
	  /**
	   * Subtracts one Z/nZ element to an other one.
	   *
	   * @param right the RingZnElement to be subtracted.
	   * @return the result of the substraction.
	   */
	  RingZnElement operator-(const RingZnElement& right) const
		  throw(RingMismatchException);


	  /**
	   * Subtracts a RingZnElement to this one.
	   *
	   * @param right a constant reference to the RingZnElement which is to
	   * be subtracted to this one.
	   */
	  void operator-=(const RingZnElement& right)
		  throw(RingMismatchException);
	  
	  /**
	   * Calculates the opposite value of this RingZnElement.
	   *
	   * @return a pointer to the opposite value of this one. It must be
	   * deleted by the caller.
	   */
	  RingZnElement operator-() const;
	  /**
	   * Multiplies two Z/nZ zlements. 
	   *
	   * @param right the right RingZnElement of the multiplication.
	   * @return the multiplication of both RingZnElement.
	   */
	  RingZnElement operator*(const RingZnElement& right) const
		  throw(RingMismatchException);

	  /**
	   * Multiplies a RingZnElement by this one.
	   *
	   * @param right a constant reference to the RingZnElement which is to
	   * be multiplied to this one.
	   */
	  void operator*=(const RingZnElement& right)
		  throw(RingMismatchException);
	  
	  /**
	   * Divides one Z/nZ element by an other one. This last element shall
	   * be invertible modulo n, that is, be coprime with n, otherwise the
	   * exception is thrown.
	   *
	   * @param denominator the denominator of the division.
	   * @throw NonInvertibleElementException<RingZnElement > if the denominator is not
	   * invertible in the ring.
	   */
	  RingZnElement operator/(const RingZnElement& denominator) const
		  throw(NonInvertibleElementException<RingZnElement>,
			RingMismatchException);

	  /**
	   * Divides this RingZnElement by an other one.
	   *
	   * @param denom a constant reference to the denominator by which this
	   * RingZnElement is divided.
	   */
	  void operator/=(const RingZnElement& denom) 
		  throw(NonInvertibleElementException<RingZnElement>,
			RingMismatchException);


	  /**
	   * Get the inverse of this RingZnElement.
	   *
	   * @return the inverse of this RingZnElement. 
	   */
	  RingZnElement inverse() const 
		  throw(NonInvertibleElementException<RingZnElement>);

	  /**
	   * Copies the value of right into this one. The value is copied (so
	   * may be destroyed in right). As usual, the ring SHALL be the same !
	   * 
	   * @param right a constant reference to the value which is to be
	   * affected.
	   * @return a reference to the results, that is the value of right.
	   */
	  RingZnElement& operator=(const RingZnElement& right);
	  
	 /**
	  * Tells whether this element is invertible or not.
	  *
	  * @return true iff this element is invertible modulo n, that is iff it
	  * is coprime with n. Return false if this element is not invertible
	  * modulo n.
	  */
	  bool isInvertible() const;

	  /**
	   * Converts the internal value of this RingZnElement to a positive one
	   * before returning it. Note that the internal value is not modified.
	   *  
	   */
	  mpz_class getPositiveValue() const;
	  
  private:
	  /**
	   * The value of this RingZnElement. It should remain, except inside
	   * computations, between -(n-1) and n-1 where n is the modulus.
	   * 
	   */
	  mpz_class* value;
	  
	  /**
	   * The RingZn this element belongs to.
	   * 
	   */
	  const RingZn * ring;

	  /**
	   * Equality operator between two RingZnElement. They must belong to
	   * the same ring.
	   *
	   * @param left the left side compared RingZnElement.
	   * @param right the right side compared RingZnElement.
	   * @return true iff left==right
	   * @throw RingMismatchException if both elements do not belong to the
	   * same ring.
	   */
	  friend bool operator==(const RingZnElement& left, const RingZnElement& right)
		  throw(RingMismatchException);
	  

 	  /**
	   * Help to debugging. Basically, sends a RingZnElement to an output stream. 
	   * ring is a pointer to the reference RingZn.
	   * 
	   * @param out a reference to the output stream where the object should be sent.
	   * @param element a constant reference to the RingZnElement which is sent.
	   * @return a reference to the resulting output stream.
	   */
	  friend std::ostream& operator<<(std::ostream& out, const RingZnElement& element);

  };

	
} //namespace libalgebra

#endif //__LIB_ALGEBRA_RING_Z_N_ELEMENT_H
